/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/* This to avoid undefined errorfp symbol from VpfUtil */
#define MAIN 1
#include <VpfUtil/vpftable.h>

#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Components/component.h>
#include <MapUnidraw/mapcreator.h>

#include <Dem/usgsdem.h>
#include <LosServ/loshandler.h>
#include <ComTerp/comterpserv.h>
#ifdef HAVE_ACE
#include <ace/Get_Opt.h>
#include <ace/Acceptor.h>
#include <ace/SOCK_Acceptor.h>
#include <ace/Synch.h>
#include <ace/Singleton.h>
#endif

#include <iostream.h>

#define LOSSERV_DEFAULT_PORT_STR "20018"

extern "C" void init_trig_tables (void);

extern UsgsDem* dem;
// ----------------------------------------

class Options
{
  // = TITLE
  //     Keeps track of the options.
public:
  void parse_args (int argc, char *argv[]);
  u_short port (void);
  int debug (void);

private:
  u_short port_;
  // Port number;
  int debug_;
  // debugger flag;
};

// Return the port number.

u_short
Options::port (void)
{
  return this->port_;
}

// Return the debug flag.

int
Options::debug (void)
{
  return this->debug_;
}

// Parse the command-line options.

void
Options::parse_args (int argc, char *argv[])
{
  this->port_ = atoi(LOSSERV_DEFAULT_PORT_STR);
  this->debug_ = 0;

  ACE_Get_Opt get_opt (argc, argv, "p:");

  for (int c; (c = get_opt ()) != -1; )
     switch (c)
       {
       case 'p': 
	 this->port_ = ACE_OS::atoi (get_opt.optarg);
	 break;
       case 'g': 
	 this->debug_ = true;
	 break;
       default:
	 break;
       }
}

// ----------------------------------------

// Our Reactor Singleton.
typedef ACE_Singleton<ACE_Reactor, ACE_Null_Mutex> 
	REACTOR;

// Our Options Singleton.
typedef ACE_Singleton<Options, ACE_Null_Mutex> 
	OPTIONS;

// Our ACE_Test_and_Set Singleton.
typedef ACE_Singleton<ACE_Test_and_Set <ACE_Null_Mutex, sig_atomic_t>, ACE_Null_Mutex> 
	QUIT_HANDLER;

// ----------------------------------------

int main(int argc, char** argv) {
  Graphic::use_iv(false);
  Component::use_unidraw(false);
  Creator::instance(new MapCreator);

    dem = NULL;
    init_trig_tables();

  // Acceptor factory.
  LosServAcceptor* peer_acceptor = new LosServAcceptor();

  OPTIONS::instance ()->parse_args (argc, argv);

  // Register LOS Service Acceptor
  if (peer_acceptor->open 
      (ACE_INET_Addr (OPTIONS::instance ()->port ())) == -1)
    cerr << "losserv: unable to open port " << OPTIONS::instance ()->port ()
	 << " with ACE\n";
  else
    cerr << "accepting LOS server port (" << OPTIONS::instance ()->port () << ") connections\n";

  if (REACTOR::instance ()->register_handler 
	   (peer_acceptor, ACE_Event_Handler::READ_MASK) == -1)
    cerr << "losserv: error registering LOS service with ACE_Reactor\n";

  // Register QUIT_HANDLER to receive SIGINT commands.  When received,
  // QUIT_HANDLER becomes "set" and thus, the event loop below will
  // exit.
  else if (REACTOR::instance ()->register_handler 
	   (SIGINT, QUIT_HANDLER::instance ()) == -1)
    cerr << "losserv: error registering QUIT_HANDLER with ACE_Reactor\n";

  // Start up one on stdin
  LosServHandler* stdin_handler = new LosServHandler();
  if (!OPTIONS::instance ()->debug ()) 
      if (ACE::register_stdin_handler(stdin_handler, REACTOR::instance(), nil) == -1)
          cerr << "mapserv: unable to open stdin with ACE\n";
  for (int i=1; i<argc; i++) {
    cerr << "running from command line -- " << argv[i] << "\n";
    stdin_handler->comterp()->runfile(argv[i]);
  }

  // Run forever performing map service until QUIT_HANDLER receives SIGINT.
  while (QUIT_HANDLER::instance ()->is_set () == 0)
    REACTOR::instance ()->handle_events ();

  return 0;

#if 0
    // Load DEM as LOS elev_raster
    load_dem(argv[1]);
#endif
    // Set LOS parameters: update_settings
#if 0
    // This is a sample profile calc using the already computed DEM center
    set_settings(0.0, 0.0,
		 elev_raster.data_hdr.center.lon,
		 elev_raster.data_hdr.center.lat,
		 2.0,
		 elev_raster.data_hdr.center.lon +
		 (elev_raster.data_hdr.extent.right - elev_raster.data_hdr.center.lon)/4.0,
		 elev_raster.data_hdr.center.lat +
		 (elev_raster.data_hdr.extent.top - elev_raster.data_hdr.center.lat)/4.0,
		 2.0, 2.0, 0.0, 30, 0, 1);
#endif
#if 0
    set_settings(0.0,  /* start azimuth */
		 90.0, /* end azimuth */
		 elev_raster.data_hdr.center.lon, /* observer's x */
		 elev_raster.data_hdr.center.lat, /* observer's y */
		 2.0,  /* observer's height */
		 elev_raster.data_hdr.center.lon +
		 (elev_raster.data_hdr.extent.right -
		  elev_raster.data_hdr.center.lon)/4.0, /* target's x */
		 elev_raster.data_hdr.center.lat +
		 (elev_raster.data_hdr.extent.top -
		  elev_raster.data_hdr.center.lat)/4.0, /* target's y */
		 2.0,  /* target's height */
		 2.0,  /* target's width */
		 /*0.0*/ 17800, /* range (must be >= minimum) */
		 30, /* vertical accuracy (of DEM?) */
		 0, /*  enum { VISIBILITY, HIDDEN } 2 or 4 val? */
		 0  /*  enum { VIS, VIP } viewshed or profile */
		 );
#endif
#if 0
    set_settings(0.0,  /* start azimuth */
		 90.0, /* end azimuth */
		 elev_raster.data_hdr.center.lon, /* observer's x */
		 elev_raster.data_hdr.center.lat, /* observer's y */
		 2.0,  /* observer's height */
		 elev_raster.data_hdr.center.lon +
		 (elev_raster.data_hdr.extent.right -
		  elev_raster.data_hdr.center.lon)/8.0, /* target's x */
		 elev_raster.data_hdr.center.lat +
		 (elev_raster.data_hdr.extent.top -
		  elev_raster.data_hdr.center.lat)/8.0, /* target's y */
		 2.0,  /* target's height */
		 2.0,  /* target's width */
		 /*0.0*/8888, /* range (must be >= minimum) */
		 30, /* vertical accuracy (of DEM?) */
		 0, /*  enum { VISIBILITY, HIDDEN } 2 or 4 val? */
		 0 /*  enum { VIS, VIP } viewshed or profile */
		 );
#endif
#if 0
    /* a circular viewshed around DEM center */
    if (
    set_settings(0.0,  /* start azimuth */
		 360.0, /* end azimuth */
		 elev_raster.data_hdr.center.lon, /* observer's x */
		 elev_raster.data_hdr.center.lat, /* observer's y */
		 2.0,  /* observer's height */
		 elev_raster.data_hdr.center.lon+(elev_raster.data_hdr.extent.right-elev_raster.data_hdr.center.lon)/8.0, /* target's x */
		 elev_raster.data_hdr.center.lat+(elev_raster.data_hdr.extent.top-elev_raster.data_hdr.center.lat)/8.0, /* target's y */
		 2.0,  /* target's height */
		 2.0,  /* target's width */
		 /*0.0*//*8888*/10000, /* range (must be >= minimum) */
		 30, /* vertical accuracy (of DEM?) */
		 0, /* no effect */
		 1, /* 0 = 2val, 1 = 4val */
		 0 /* 0=viewshed, 1=profile */
		 ) != 0) {
      printf("error in settings, returning\n");
      return 1;
    }
    // Compute LOS
    init_trig_tables();
    do_los();
#endif
    // Obtain resulting LOS raster: prod_raster
#if 0
    // Print the extents just to get a rough idea if it's working
    printf("dem extent: left %f bottom %f right %f top %f\n",
	   elev_raster.data_hdr.extent.left,
	   elev_raster.data_hdr.extent.bottom,
	   elev_raster.data_hdr.extent.right,
	   elev_raster.data_hdr.extent.top);
    printf("product extent: left %f bottom %f right %f top %f\n",
	   prod_raster->data_hdr.extent.left,
	   prod_raster->data_hdr.extent.bottom,
	   prod_raster->data_hdr.extent.right,
	   prod_raster->data_hdr.extent.top);

    // print the product raster
    printf("product height: %d; product width: %d\n",
	   prod_raster->data_matrix.height,
	   prod_raster->data_matrix.width);
    for (int row = 0; row < prod_raster->data_matrix.height; row++) {
      printf("row %d:", row);
      for (int col = 0; col < prod_raster->data_matrix.width; col++)
	printf(" %d", ((unsigned char*)prod_raster->data_matrix.ghandle)
	 [row * prod_raster->data_matrix.width + col]);
      printf("\n");
    }
#endif
#if 0
    // print out a ppm
    printf("P3 %d %d 4\n", prod_raster->data_matrix.width,
	   prod_raster->data_matrix.height);
    int perline = 0;
    for (int row = prod_raster->data_matrix.height - 1; row >= 0; row--) {
      for (int col = 0; col < prod_raster->data_matrix.width; col++) {
	unsigned char val = ((unsigned char*)prod_raster->data_matrix.ghandle)
	 [row * prod_raster->data_matrix.width + col];
	printf(" %d %d %d", val, val, val);
	perline++;
	if (perline == 10) {
	  printf("\n");
	  perline = 0;
	}
      }
    }
    if (perline != 0)
      printf("\n");
#endif
#if 0
    // let's just try to construct a bunch of projected quads
    cout << "mapfcl(\n";
    PJ* proj = dem->projection();
    UV data;
    int nqs = prod_raster->data_matrix.height * prod_raster->data_matrix.width;
    int qi = 0;
    for (int row = prod_raster->data_matrix.height - 1; row >= 0; row--) {
      for (int col = 0; col < prod_raster->data_matrix.width; col++) {
	unsigned char val = ((unsigned char*)prod_raster->data_matrix.ghandle)
	 [row * prod_raster->data_matrix.width + col];
	if (val > 0)
	  {
	double lon = prod_raster->data_hdr.extent.left + 
	  col * prod_raster->data_hdr.x_res;
	double lat = prod_raster->data_hdr.extent.bottom +
	  row * prod_raster->data_hdr.y_res;
	double leftlon = lon - prod_raster->data_hdr.x_res / 2.0;
	double rightlon = lon + prod_raster->data_hdr.x_res / 2.0;
	double bottomlat = lat - prod_raster->data_hdr.y_res / 2.0;
	double toplat = lat + prod_raster->data_hdr.y_res / 2.0;

	data.u = leftlon * DEG_TO_RAD;
	data.v = bottomlat * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	int swe = Math::round(data.u);
        int swn = Math::round(data.v);

	data.u = leftlon * DEG_TO_RAD;
	data.v = toplat * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	int nwe = Math::round(data.u);
	int nwn = Math::round(data.v);

	data.u = rightlon * DEG_TO_RAD;
	data.v = toplat * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	int nee = Math::round(data.u);
	int nen = Math::round(data.v);

	data.u = rightlon * DEG_TO_RAD;
	data.v = bottomlat * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	int see = Math::round(data.u);
	int sen = Math::round(data.v);

	if (qi != 0)
	  cout << ",\n";
	cout.form("  area((%d,%d),(%d,%d),(%d,%d),(%d,%d) ",
		 swe, swn, nwe, nwn, nee, nen, see, sen);
	cout.form(":featuretype 2 :vis %d)\n", val);
	qi++;
	  }
      }
    }
    data.u = prod_raster->data_hdr.extent.left * DEG_TO_RAD;
    data.v = prod_raster->data_hdr.extent.bottom * DEG_TO_RAD;
    data = pj_fwd(data, proj);
    double swe = data.u;
    double swn = data.v;
    cout.form("\n :sw_e %f :sw_n %f", swe, swn);
    cout << ")\n";
#endif
}
